package dev.ninjdai.werewolf.worldgen;

import net.minestom.server.coordinate.Point;
import net.minestom.server.coordinate.Pos;
import net.minestom.server.coordinate.Vec;
import net.minestom.server.instance.block.Block;
import net.minestom.server.instance.generator.GenerationUnit;
import net.minestom.server.instance.generator.Generator;
import net.minestom.server.world.biome.Biome;
import net.minestom.vanilla.generation.VanillaWorldGenerationFeature;
import net.minestom.vanilla.tag.Tags;
import org.jetbrains.annotations.NotNull;

import java.util.Random;

public class LGGenerator implements Generator {
    TerrainBuilder terrainBuilder;
    Random random;
    double densityThreshold = 1;//0.6;
    int upperRelativeDensityLimit = 50;
    int lowerRelativeDensityLimit = 50;

    public LGGenerator(TerrainBuilder tb) {
        this.terrainBuilder = tb;
        this.random = new Random(tb.seed);
    }

    @Override
    public void generate(@NotNull GenerationUnit unit) {
        final Point min = unit.absoluteStart();
        final Point max = unit.absoluteEnd();

        for (int x = min.blockX(); x < max.blockX(); x++) {
            for (int z = min.blockZ(); z < max.blockZ(); z++) {
                generate3D(min, max, x, z, unit);
            }
        }
    }

    private void generate3D(Point min, Point max, int x, int z, GenerationUnit unit) {
        int height = terrainBuilder.getSurfaceHeight(x, z);
        int upperDensityLimit = height + upperRelativeDensityLimit;
        int lowerDensityLimit = height - lowerRelativeDensityLimit;
        int terrainHeight = lowerDensityLimit;

        for (int y = upperDensityLimit; y >= lowerDensityLimit; y--) {
            if (y == lowerDensityLimit) {
                unit.modifier().fill(min.withX(x).withZ(z), new Vec(x + 1, lowerDensityLimit + 1, z + 1), Block.STONE);
                break;
            }

            Vec pos = new Vec(x, y, z);
            if (terrainBuilder.getDensity(x, y, z) > densityThreshold) {
                unit.modifier().setBlock(pos, Block.STONE);
                terrainHeight = Math.max(terrainHeight, y);
            }
        }
        decorate(x, terrainHeight, z, unit);
    }

    private void decorate(int x, int y, int z, GenerationUnit unit) {
        unit.modifier().setBlock(x, y + 4, z, Block.GRASS_BLOCK);
        for (int i = 0; i < 4; i++) {
            Block b = switch ((int) (2*Noises.FEATURES.evaluateNoise(x, y+i, z))) {
                case 0 -> Block.COARSE_DIRT;
                case 1 -> Block.ROOTED_DIRT;
                default -> Block.DIRT;
            };
            unit.modifier().setBlock(x, y + i, z, b);
        }
        int ry = y+5;
        double ft = Noises.FEATURES.evaluateNoise(x, z);
        if (ft > 0.995) {
            int height = (int) (3 + 3*Noises.FEATURES.evaluateNoise(x, y, z));

            GenerationUnit fork = unit.fork(new Vec(x-2, ry, z-2), new Vec(x+3, ry+height+4, z+3));
            fork.modifier().fill(new Vec(x, ry, z), new Vec(x+1, ry+height, z+1), Block.OAK_LOG);
            fork.modifier().fill(new Vec(x-2, ry+height, z-2), new Vec(x+3, ry+height+2, z+3), Block.OAK_LEAVES);
            fork.modifier().fill(new Vec(x-1, ry+height+2, z-1), new Vec(x+2, ry+height+3, z+2), Block.OAK_LEAVES);
        } else if (ft > 0.95) {
            unit.modifier().setBlock(x, ry, z, Block.TALL_GRASS);
            unit.modifier().setBlock(x, ry+1, z, Block.TALL_GRASS.withProperty("half", "upper"));
        } else if (ft > 0.8) {
            unit.modifier().setBlock(x, ry, z, Block.SHORT_GRASS);
        } else if (ft < -0.99) {
            unit.modifier().setBlock(x, ry, z, Block.POPPY);
        }
    }
}